#!/bin/bash
########################################################################
# build_triton.sh
#
# Build triton and package for UPS.
#
# Adapted from build.sh.template provided by ssibuildshims v2_09_00.
########################################################################

########################################################################
# Customization instructions.
#
# 1. Replace triton with the UPS name of your product (e.g., sqlite).
# 2. Replace v2_3_0 with the product's UPS version string.
# 3. Customize all appropriate sections or commands marked with *C*.
# 4. Remove inapplicable sections and comments---including this
#    block---from your build script.
#
########################################################################

########################################################################
# Non-customize-able boilerplate: skip to "Customize-able functions."
########################################################################

####################################
# Useful variables.
prog=${BASH_SOURCE[0]##*/}
thisdir=$(cd "${BASH_SOURCE%/*}" && pwd -P)

##################
# Local (dumb) ssi_die() function for use until we source ssi_functions.
ssi_die() {
  printf "FATAL ERROR: ${*}\n" 1>&2; exit 1
}
##################

##################
# Basic usage function for use by ensure_buildshims until we have enough
# information to provide something more specific. Do NOT customize!
build_script_usage() {
    cat <<EOF
USAGE: ${prog} <ssi-msg-config-options> [--] <product-dir> <qualifier-args> <maketar>
       ${prog} -h|--help [<product-dir>]

  Build ${package} and package for UPS.

OPTIONS

  -h|--help

    Help. Supply <product-dir> for specific, detailed information on
    invoking ${prog}.

  <ssi-msg-config-options>

    Supply <product-dir> for more information.

ARGUMENTS

  <product-dir>

    The top-level UPS directory in which the source code distribution
    was unpacked. When specified with -h|--help, this enables us to
    provide more specific help about invoking ${prog}.

  <qualifier-args>
  <maketar>

    Supply <product-dir> for more information.

EOF
}
##################

##################
# Make sure we have UPS and ssibuildshims set up.
ensure_ssibuildshims() {
  if ! (( $# )) || [ "${@: -1}" = "-h" ] || [ "${@: -1}" = "--help" ]; then
    build_script_usage
    exit 1
  fi
  # Find <product-dir> from our arguments:
  local arg
  for arg in "${@}"; do
    [[ "$arg" ==  -* ]] && continue;
    product_dir="${arg}"; break
  done
  if [ -z "${product_dir}" ]; then
    # Bad usage.
    ssi_die "Bad <product-dir> ${product_dir}\n$(build_script_usage)"
    exit 2
  elif ! { [ -d "${product_dir}/.upsfiles" ] && [ -w "${product_dir}" ]; }; then
    # Something wrong with <product-dir>.
    ssi_die "<product-dir> not a valid, writable unified UPS directory: " \
      "${product_dir}\n$(build_script_usage)"
  fi
  # Promising...
  local abs_product_dir="$(cd "${product_dir}" 2>/dev/null 2>&1 && pwd -P)"
  (( $? )) && ssi_die "unable to cd to <product-dir>: ${product_dir}\n" \
    "$(build_script_usage)"
  # Find our own bootstraps and yank 'em.
  local ssibuildshims_topdir="${product_dir}/ssibuildshims/${ssibuildshims_version}"
  [ -f "${ssibuildshims_topdir}/bin/ssi_functions" ] && \
    source "${ssibuildshims_topdir}/bin/ssi_functions" >/dev/null 2>&1 && \
    ensure_ups "${product_dir}" && \
    setup ssibuildshims ${ssibuildshims_version} -z "${product_dir}" || \
    ssi_die "unable to find ssibuildshims ${ssibuildshims_version} in
${product_dir} and set it up with UPS."
  # Absolute is safer.
  absolute_path -- "${product_dir}" product_dir
}
##################

########################################################################
# Customize-able functions.
########################################################################

##################
# Ascertain whether we built and installed correctly. Add more /
# different calls to verify_in_filesystem as appropriate. Arguments
# (e.g. -q) are passed through to verify_in_filesystem.
is_built() { # *C*
  ${trace} verify_in_filesystem -d "${@}" "${pkgdir}/lib"{64,}
}
##################

##################
# If a distribution archive is wanted and the installation is successful
# according to is_built(), then make it. Any arguments (e.g. -q) are
# passed through both to is_built() and make_basic_distribution().
#
# If make_basic_distribution() finds that the product is declared
# correctly and attempts to make the distribution, then it will cause us
# to exit with the status code of that operation. If not, it returns an
# error code.
#
# See make_basic_distribution -h for useful options to add to
# mbd_opts(). *C*
maybe_make_distribution() {
  if [ "${maketar}" = "tar" ]; then
    # Customize mbd_opts. *C*
    local mbd_opts=()
    ${trace} is_built "$@" && \
      ${trace} make_basic_distribution "${@}" "${mbd_opts[@]}"
  fi
}
##################

########################################################################
# Package-specific functions.
########################################################################

# ...

########################################################################
# Main.
########################################################################

####################################
# Useful variables. *C*
trace=maybe_xtrace # Could set to xtrace instead for unconditional.
(( ssi_trace = 1 )) # Trace certain commands by default.
####################################

####################################
# Package name and version information. *C*
. ${thisdir}/version_info.sh || \
  ssi_die "unable to load version info from ${thisdir}/version_info"
####################################

####################################
# Handle options and arguments.

# Set up ssibuildshims and UPS / product_dir as early as possible.
ensure_ssibuildshims "${@}"

##################
# Now parse options and arguments, setting basequal, extraqual, maketar,
# etc.
#
# Use verify_quals() to parse command line options, verify that the user
# has provided qualifiers supported by this build script, and set
# expected variables such as basequal, extraqual, cqual, etc.
#
# Customize the extra usage text as appropriate. Note that DEBUG_PATCHES
# and VERBOSE_PATCHES are honored by prepare_source() and
# patch_source(). DEBUG_(CONFIG|BUILD|TEST|INSTALL) should be accounted
# for (or not) by the relevant command in this script.
#
# See verify_quals -h for options and examples.
verify_quals \
  --cqual={e19,e20,c7} \
  --parse --usage-file=- -- "$@" <<EOF || exit # *C*
ENVIRONMENT

  DEBUG_(CONFIG|BUILD|TEST|INSTALL)

    Verbose output from the corresponding step.

  DEBUG_PATCHES
  VERBOSE_PATCHES

    Command tracing and increased verbosity for source unpacking and patching
    operations. DEBUG_PATCHES will cause exit after patches have been applied.
EOF
##################

##################
# Define globals: pkgdir, sourcedir, blddir, fullqual, etc. See
# define_basics -h for steering options.
${trace} define_basics NO_TYPE TRACK_PROGRESS || exit # *C*
##################

####################################
# Declare ourselves privately, and then set ourselves up to allow setup
# of dependencies. *C*
#
# Comment or delete if not required:
${trace} fake_declare_product "${ssi_args[@]}" -x -- || exit
####################################

####################################
# Build-only dependencies. *C*
#
# e.g.
   ${trace} setup_cmake --latest={--ge,3.18} || exit
#      (version requirement necessary for cmake --install).
# AND/OR
#   ${trace} setup_ninja --optional && (( want_ninja = 1 ))
####################################

# Print active products.
echo
ups active
echo

# Possible short-circuit if we're already built and declared.
${trace} maybe_make_distribution --silent

ssi_info "building ${package} for ${OS}-${plat}-${qualdir} (flavor ${pkgflvr})"

# Ensure directories: pkgdir, sourcedir, blddir.
${trace} init_pkg_dirs || exit

####################################
# Prepare source for building. *C*

${trace} cd "${sourcedir}" \
  || ssi_die "unable to cd to source directory ${sourcedir}"

patches=( cms.patch )
# Source tarballs and patches. See prepare_source -h / patch_source -h
# for options.
${trace} prepare_source "${ssi_args[@]}" -P "${patchdir}" \
  "${tardir}/${pkgtarfile}" "${patches[@]}" || exit

# Bail early if DEBUG_PATCHES is set.
[ -n "${DEBUG_PATCHES}" ] && exit 1
####################################

####################################
# Prepare to build. *C*
#
# Set configure / compile flags, environment variables, etc.

# double checking
if [ -z ${PROTOBUF_FQ_DIR} ]
then
   echo "ERROR: failed to setup protobuf"
   exit 1
else
   echo "protobuf is in ${PROTOBUF_FQ_DIR}"
fi
if [ -z ${OPENCV_FQ_DIR} ]
then
   echo "ERROR: failed to setup protobuf"
   exit 1
else
   echo "opencv is in ${OPENCV_FQ_DIR}"
fi
if [ -z ${PYTHON_DIR} ]
then
    echo "ERROR: failed to setup python"
    exit 1
else
    echo "python is in ${PYTHON_DIR}"
fi

std=17
cxxflg="-Wno-deprecated-declarations"
case ${basequal} in
  e19|e20) cc=gcc; cxx=g++;;
  c7) cc=clang; cxx=clang++;;
  *) ssi_die "Qualifier $basequal not recognized."
esac

# Require several python facilities
python_shortdotversion=${PYTHON_VERSION#v}
python_shortdotversion=${python_shortdotversion//_/.}
python_shortdotversion=${python_shortdotversion%.*}

pip install --prefix $pkgdir wheel setuptools grpcio-tools
export PYTHONPATH=${pkgdir}/lib/python${python_shortdotversion}/site-packages

# ...

##################
# Configure parallelism. *C*
ncores=$(ncores)

# For cmake --build:
#
# [[ -n "${CMAKE_BUILD_PARALLEL_LEVEL}" ]] || \
#   export CMAKE_BUILD_PARALLEL_LEVEL=${ncores}
#
# For ctest:
#
# [[ -n "${CTEST_PARALLEL_LEVEL}" ]] || \
#   export CTEST_PARALLEL_LEVEL=${ncores}
##################
####################################

####################################
# Configure, build, and install.
#
# Use ${blddir}/${pkgtarname} for in-source builds, ${blddir} otherwise.

${trace} cd "${blddir}" || \
  ssi_die "unable to cd to build directory ${blddir}"

##################
# Configure. *C*

# e.g.
#   ${trace} ${sourcedir}/${pkgtarname}/configure \
#     --prefix="${pkgdir}" ... || ssi_die "configure failed"
# OR
#   ${trace} cmake <opts> ${DEBUG_CONFIG:+--verbose} \
#     "${sourcedir}/${pkgtarname}" || \
#     ssi_die "configure failed"

${trace} cmake ${sourcedir}/${pkgtarname}/build/client \
      -DCMAKE_INSTALL_PREFIX=${pkgdir} \
      -DCMAKE_BUILD_TYPE=MinSizeRel \
      -DCMAKE_C_COMPILER=${cc} \
      -DCMAKE_CXX_COMPILER=${cxx} \
      -DCMAKE_CXX_FLAGS="${cxxflg} -Wno-error=deprecated-declarations" \
      -DTRITON_CLIENT_SKIP_EXAMPLES=ON \
      -DTRITON_ENABLE_HTTP=OFF \
      -DTRITON_ENABLE_GRPC=ON \
      -DTRITON_ENABLE_GPU=OFF \
      -DProtobuf_DIR=${PROTOBUF_FQ_DIR}/lib/cmake/protobuf \
      -DTRITON_VERSION=${pkgdotver}

##################

##################
# Build. *C*

# e.g.
#   ${trace} cmake --build "${blddir}" ${DEBUG_BUILD:+--verbose} || \
#     ssi_die "build failed"
${trace} make -j ${ncores}  || ssi_die "make client failed"
##################

##################
# Test. *C*

# e.g.
# (( WANT_TESTS )) && \
#   { ${trace} ctest --output-on-failure || ssi_die "tests failed"; }
##################

##################
# Install. *C*

# e.g.
#   ${trace} cmake --install "${blddir}" || ssi_die "install failed"
${trace} make -j ${ncores} install  || ssi_die "make install failed"
##################

ssi_info "finished building ${package} ${pkgver}"

####################################
# Pre-declaration processing. *C*
#
# Move things, install from contrib, clean build products from in-source
# builds, etc.
####################################

##################
# Declare and set up the installed package. *C*
#
# Add -c for current declaration if required.
${trace} declare_product -x -- || exit
##################

####################################
# Post-declaration processing. *C*
#
# Operations requiring the *final* declaration and setup (as opposed to
# the pre-build fake declaration and setup). This should be rare.
####################################

#################
# Verify installation and make distribution archive. *C*
#
# To customize verification, see is_built(), since it is also used by
# maybe_make_distribution(). See the documentation for those functions
# (above) for customization opportunities.

# Add a message to ssi_build_failed if appropriate.
${trace} is_built && ${trace} ssi_build_completed && \
  ssi_info "${package} is installed at ${pkgdir}" && \
  ${trace} maybe_make_distribution || \
  ssi_build_failed # *C*
########################################################################
# Nothing beyond this point.
########################################################################
